@@ -17,8 +17,11 @@ from torch import nn, cat, stack, arange, tensor, Tensor, is_tensor, zeros, ones
 import torchvision
 from torchvision.models import VGG16_Weights
 
-from x_mlps_pytorch.normed_mlp import create_mlp
+from torch.optim import Optimizer
+from adam_atan2_pytorch import MuonAdamAtan2
+
 from x_mlps_pytorch.ensemble import Ensemble
+from x_mlps_pytorch.normed_mlp import create_mlp
 
 from hyper_connections import get_init_and_expand_reduce_stream_functions
 
@@ -38,6 +41,7 @@ from assoc_scan import AssocScan
 # g - groups of query heads to key heads (gqa)
 # vc - video channels
 # vh, vw - video height and width
+# mtp - multi token prediction length
 
 import einx
 from einx import add, multiply
@@ -65,7 +69,7 @@ TokenizerLosses = namedtuple('TokenizerLosses', ('recon', 'lpips'))
 WorldModelLosses = namedtuple('WorldModelLosses', ('flow', 'reward', 'behavior_clone'))
 
 @dataclass
-class WorldModelGenerations:
+class Experience:
     latents: Tensor
     video: Tensor | None = None
     rewards: Tensor | None = None
@@ -74,6 +78,7 @@ class WorldModelGenerations:
     values: Tensor | None = None
     step_size: int | None = None
     agent_index: int = 0
+    is_from_world_model: bool = True
 
 # helpers
 
@@ -170,6 +175,30 @@ def l2norm(t):
 def softclamp(t, value = 50.):
     return (t / value).tanh() * value
 
+def create_multi_token_prediction_targets(
+    t, # (b t ...)
+    steps_future,
+
+): # (b t-1 steps ...), (b t-1 steps) - targets and the mask, where mask is False for padding
+
+    batch, seq_len, device = *t.shape[:2], t.device
+
+    batch_arange = arange(batch, device = device)
+    seq_arange = arange(seq_len, device = device)[1:]
+    steps_arange = arange(steps_future, device = device)
+
+    indices = add('t, steps -> t steps', seq_arange, steps_arange)
+    mask = indices < seq_len
+
+    batch_arange = rearrange(batch_arange, 'b -> b 1 1')
+
+    indices[~mask] = 0
+    mask = repeat(mask, 't steps -> b t steps', b = batch)
+
+    out = t[batch_arange, indices]
+
+    return out, mask
+
 # loss related
 
 class LPIPSLoss(Module):
@@ -505,7 +534,8 @@ class ActionEmbedder(Module):
         continuous_targets = None,       # (... na)
         discrete_action_types = None,    # (na)
         continuous_action_types = None,  # (na)
-        parallel_discrete_calc = None
+        parallel_discrete_calc = None,
+        return_entropies = False
     ):
         parallel_discrete_calc = default(parallel_discrete_calc, exists(discrete_targets) and discrete_targets.shape[-1] > 1)
 
@@ -514,6 +544,7 @@ class ActionEmbedder(Module):
         # discrete
 
         discrete_log_probs = None
+        discrete_entropies = None
 
         if exists(discrete_targets):
 
@@ -533,7 +564,18 @@ class ActionEmbedder(Module):
 
                 nested_logits = nested_tensor(discrete_action_logits.split(jagged_dims * batch), layout = torch.jagged, device = self.device, requires_grad = True)
 
-                log_probs = log(nested_logits.softmax(dim = -1))
+                prob = nested_logits.softmax(dim = -1)
+
+                log_probs = log(prob)
+
+                # maybe entropy
+
+                if return_entropies:
+                    discrete_entropies = (-prob * log_probs).sum(dim = -1, keepdim = True)
+                    discrete_entropies = cat(discrete_entropies.unbind())
+                    discrete_entropies = rearrange(discrete_entropies, '(b na) -> b na', b = batch)
+
+                    discrete_entropies = inverse_pack_lead_dims(discrete_entropies, '* na')
 
                 # back to regular tensor
 
@@ -557,20 +599,30 @@ class ActionEmbedder(Module):
 
             else:
                 discrete_log_probs = []
+                discrete_entropies = []
 
                 for one_discrete_action_logit, one_discrete_target in zip(discrete_action_logits, discrete_targets.unbind(dim = -1)):
 
-                    one_discrete_log_probs = one_discrete_action_logit.log_softmax(dim = -1)
+                    one_discrete_probs = one_discrete_action_logit.softmax(dim = -1)
+                    one_discrete_log_probs = log(one_discrete_probs)
                     one_discrete_target = rearrange(one_discrete_target, '... -> ... 1')
 
                     log_prob = one_discrete_log_probs.gather(-1, one_discrete_target)
                     discrete_log_probs.append(log_prob)
 
+                    if return_entropies:
+                        entropy = (-one_discrete_probs * one_discrete_log_probs).sum(dim = -1)
+                        discrete_entropies.append(entropy)
+
                 discrete_log_probs = cat(discrete_log_probs, dim = -1)
 
+                if return_entropies:
+                    discrete_entropies = stack(discrete_entropies, dim = -1)
+
         # continuous
 
         continuous_log_probs = None
+        continuous_entropies = None
 
         if exists(continuous_targets):
             mean, log_var = continuous_action_mean_log_var.unbind(dim = -1)
@@ -579,7 +631,17 @@ class ActionEmbedder(Module):
             distr = Normal(mean, std)
             continuous_log_probs = distr.log_prob(continuous_targets)
 
-        return discrete_log_probs, continuous_log_probs
+            if return_entropies:
+                continuous_entropies = distr.entropy()
+
+        log_probs = (discrete_log_probs, continuous_log_probs)
+
+        if not return_entropies:
+            return log_probs
+
+        entropies = (discrete_entropies, continuous_entropies)
+
+        return log_probs, entropies
 
     def forward(
         self,
@@ -1457,6 +1519,7 @@ class DynamicsWorldModel(Module):
         num_continuous_actions = 0,
         continuous_norm_stats = None,
         reward_loss_weight = 0.1,
+        multi_token_pred_len = 8,
         value_head_mlp_depth = 3,
         policy_head_mlp_depth = 3,
         behavior_clone_weight = 0.1,
@@ -1584,6 +1647,10 @@ class DynamicsWorldModel(Module):
 
         self.behavior_clone_weight = behavior_clone_weight
 
+        # multi token prediction length
+
+        self.multi_token_pred_len = multi_token_pred_len
+
         # each agent token will have the reward embedding of the previous time step - but could eventually just give reward its own token
 
         self.add_reward_embed_to_agent_token = add_reward_embed_to_agent_token
@@ -1595,11 +1662,16 @@ class DynamicsWorldModel(Module):
             learned_embedding = add_reward_embed_to_agent_token
         )
 
-        self.to_reward_pred = Sequential(
+        to_reward_pred = Sequential(
             RMSNorm(dim),
             LinearNoBias(dim, self.reward_encoder.num_bins)
         )
 
+        self.to_reward_pred = Ensemble(
+            to_reward_pred,
+            multi_token_pred_len
+        )
+
         self.reward_loss_weight = reward_loss_weight
 
         # value head
@@ -1663,18 +1735,20 @@ class DynamicsWorldModel(Module):
 
         return list(set(params) - set(self.video_tokenizer.parameters()))
 
-    def learn_policy_from_generations(
+    def learn_from_experience(
         self,
-        generation: WorldModelGenerations
+        experience: Experience,
+        policy_optim: Optimizer | None = None,
+        value_optim: Optimizer | None = None
     ):
-        latents = generation.latents
-        actions = generation.actions
-        old_log_probs = generation.log_probs
-        old_values = generation.values
-        rewards = generation.rewards
+        latents = experience.latents
+        actions = experience.actions
+        old_log_probs = experience.log_probs
+        old_values = experience.values
+        rewards = experience.rewards
 
-        step_size = generation.step_size
-        agent_index = generation.agent_index
+        step_size = experience.step_size
+        agent_index = experience.agent_index
 
         assert all([*map(exists, (old_log_probs, actions, old_values, rewards, step_size))]), 'the generations need to contain the log probs, values, and rewards for policy optimization'
 
@@ -1707,12 +1781,13 @@ class DynamicsWorldModel(Module):
 
         policy_embed = self.policy_head(agent_embed)
 
-        log_probs = self.action_embedder.log_probs(policy_embed, discrete_targets = discrete_actions, continuous_targets = continuous_actions)
+        log_probs, entropies = self.action_embedder.log_probs(policy_embed, discrete_targets = discrete_actions, continuous_targets = continuous_actions, return_entropies = True)
 
         # concat discrete and continuous actions into one for optimizing
 
         old_log_probs = safe_cat(old_log_probs, dim = -1)
         log_probs = safe_cat(log_probs, dim = -1)
+        entropies = safe_cat(entropies, dim = -1)
 
         ratio = (log_probs - old_log_probs).exp()
 
@@ -1723,10 +1798,27 @@ class DynamicsWorldModel(Module):
         # clipped surrogate loss
 
         policy_loss = -torch.min(ratio * advantage, clipped_ratio * advantage)
-
         policy_loss = reduce(policy_loss, 'b t na -> b t', 'sum')
+
         policy_loss = policy_loss.mean()
 
+        # handle entropy loss for naive exploration bonus
+
+        entropy_loss = - reduce(entropies, 'b t na -> b t', 'sum').mean()
+
+        total_policy_loss = (
+            policy_loss +
+            entropy_loss * self.policy_entropy_weight
+        )
+
+        # maye take policy optimizer step
+
+        if exists(policy_optim):
+            total_policy_loss.backward()
+
+            policy_optim.step()
+            policy_optim.zero_grad()
+
         # value loss
 
         value_bins = self.value_head(agent_embed)
@@ -1742,7 +1834,15 @@ class DynamicsWorldModel(Module):
 
         value_loss = torch.maximum(value_loss_1, value_loss_2).mean()
 
-        return policy_loss, value_loss
+        # maybe take value optimizer step
+
+        if exists(policy_optim):
+            value_loss.backward()
+
+            value_optim.step()
+            value_optim.zero_grad()
+
+        return total_policy_loss, value_loss
 
     @torch.no_grad()
     def generate(
@@ -1751,6 +1851,7 @@ class DynamicsWorldModel(Module):
         num_steps = 4,
         batch_size = 1,
         agent_index = 0,
+        tasks: int | Tensor | None = None,
         image_height = None,
         image_width = None,
         return_decoded_video = None,
@@ -1764,9 +1865,18 @@ class DynamicsWorldModel(Module):
         was_training = self.training
         self.eval()
 
+        # validation
+
         assert log2(num_steps).is_integer(), f'number of steps {num_steps} must be a power of 2'
         assert 0 < num_steps <= self.max_steps, f'number of steps {num_steps} must be between 0 and {self.max_steps}'
 
+        if isinstance(tasks, int):
+            tasks = torch.full((batch_size,), tasks, device = self.device)
+
+        assert not exists(tasks) or tasks.shape[0] == batch_size
+
+        # get state latent shape
+
         latent_shape = self.latent_shape
 
         # derive step size
@@ -1819,6 +1929,7 @@ class DynamicsWorldModel(Module):
                     signal_levels = signal_levels_with_context,
                     step_sizes = step_size,
                     rewards = decoded_rewards,
+                    tasks = tasks,
                     discrete_actions = decoded_discrete_actions,
                     continuous_actions = decoded_continuous_actions,
                     latent_is_noised = True,
@@ -1847,7 +1958,7 @@ class DynamicsWorldModel(Module):
             if return_rewards_per_frame:
                 one_agent_embed = agent_embed[:, -1:, agent_index]
 
-                reward_logits = self.to_reward_pred(one_agent_embed)
+                reward_logits = self.to_reward_pred.forward_one(one_agent_embed, id = 0)
                 pred_reward = self.reward_encoder.bins_to_scalar_value(reward_logits, normalize = True)
 
                 decoded_rewards = cat((decoded_rewards, pred_reward), dim = 1)
@@ -1914,11 +2025,12 @@ class DynamicsWorldModel(Module):
 
         # returning agent actions, rewards, and log probs + values for policy optimization
 
-        gen = WorldModelGenerations(
+        gen = Experience(
             latents = latents,
             video = video,
             step_size = step_size,
-            agent_index = agent_index
+            agent_index = agent_index,
+            is_from_world_model = True
         )
 
         if return_rewards_per_frame:
@@ -2282,8 +2394,19 @@ class DynamicsWorldModel(Module):
             if rewards.ndim == 2: # (b t)
                 encoded_agent_tokens = reduce(encoded_agent_tokens, 'b t g d -> b t d', 'mean')
 
-            reward_pred = self.to_reward_pred(encoded_agent_tokens)
-            reward_loss = F.cross_entropy(reward_pred, two_hot_encoding)
+            reward_pred = self.to_reward_pred(encoded_agent_tokens[:, :-1])
+
+            reward_pred = rearrange(reward_pred, 'mtp b t l -> b l t mtp')
+
+            reward_targets, reward_loss_mask = create_multi_token_prediction_targets(two_hot_encoding, self.multi_token_pred_len)
+
+            reward_targets = rearrange(reward_targets, 'b t mtp l -> b l t mtp')
+
+            reward_losses = F.cross_entropy(reward_pred, reward_targets, reduction = 'none')
+
+            reward_losses = reward_losses.masked_fill(reward_loss_mask, 0.)
+
+            reward_loss = reward_losses.sum(dim = -1).mean() # they sum across the prediction steps - eq(9)
 
         # maybe autoregressive action loss
 
@@ -2308,10 +2431,10 @@ class DynamicsWorldModel(Module):
             )
 
             if exists(discrete_log_probs):
-                behavior_clone_loss = behavior_clone_loss + discrete_log_probs.sum(dim = -1).mean()
+                behavior_clone_loss = behavior_clone_loss - discrete_log_probs.sum(dim = -1).mean()
 
             if exists(continuous_log_probs):
-                behavior_clone_loss = behavior_clone_loss + continuous_log_probs.sum(dim = -1).mean()
+                behavior_clone_loss = behavior_clone_loss - continuous_log_probs.sum(dim = -1).mean()
 
         # gather losses
 
